#include "simpleDB_SD.h"

/** init an new or existing SD datanase **/
DB_RESULT simpleDB::initDB(const char *nameDB, const char *backupDB, uint8_t ssPin, uint8_t keyLen, uint8_t recLen, uint16_t maxRec) {
  // init DB parameters
  strcpy(dbN, nameDB);    // database file name    //  max len 31
  strcpy(cpN, backupDB);  // copy/backup file name
  ssP = ssPin;       // Chip Select pin
  kyW = keyLen;      // key width 
  dbW = recLen;      // record width, key|t|dddd|...|ddd (t=tombstone, ddd..=data, |=delimiter)
  mxR = maxRec;      // max number of records
  tsM = kyW + 1;     // pos. tombstone marker
  nxW = kyW + 3;     // index width (key width + tombstone marker "key|0|")
  dtW = dbW - nxW;   // max data width in a record, exclusive 'key' and 'tombstone' marker
  dbB = new char[dbW],      // record read buffer
  nxB = new char[nxW + 1];  // index read buffer

  // used in addRec(): create a strcpy() format string (fixed record size)
  strcpy(hlK,("%" + String(kyW) + "s|0|").c_str());  // if kwW =   4 then hlK = "%4s|0|",  "|" is delimiter, "0" if tombstone
  strcpy(hlR,("%-" + String(dtW) + "s").c_str());    // if dtW = 104 then hlR = "%-104s"
  if (!SD.begin(ssPin)) {dbResult = MOUNT_ERR; return dbResult;}
  openDB();  // SD start,
  return dbResult;
}

/** open db **/
DB_RESULT simpleDB::openDB() {
  dbResult = DB_OK;
  if (dbF && (dbR == 0)) dbResult = nrRec();
  else {
   dbF = SD.open(dbN, FILE_WRITE);
   if (!dbF) dbResult = OPEN_ERR;
   else if (dbR == 0) dbResult = nrRec();
  }
  return dbResult;
}

/** close db **/
void simpleDB::closeDB() {
  dbF.close();
}

/** set counter dbR, the nummer of records, tombstone records are excluded **/
DB_RESULT simpleDB::nrRec() {  // initDB() call this function, simpleDB update dbR it and you can use dbR
  dbResult = DB_OK;
  uint16_t len;
  dbR = dbT = 0;
  if (dbF.size() < dbW) return dbResult;  // db is empty
  if (!dbF.seek(0)) return dbResult;
  while (dbF.available()) {
   len = dbF.readBytes(nxB, nxW);
   nxB[len] = '\0';
   if (len != nxW) {dbResult = CORRUPT; break;}
   if (nxB[tsM] == '0') dbR++; // it's not a tombstone
   dbT++;
   if (!dbF.seek(dbF.position() + dtW)) break;  // skip data, set entree to next record
  }
  return dbResult;
}

/** add a record with a UNIQUE key **/
DB_RESULT simpleDB::addRec(const char *key, const char *data) {  // key UNIQUE!!
  if (openDB() != DB_OK) return dbResult;
  if (dbR == mxR) {dbResult = MAX_REC; return dbResult;}  // max records in db
  if (strlen(data) > dtW) {dbResult = MAX_CHARS; return dbResult;}  // data: too many characters
  if (findRec(key) == REC_EXIST) return REC_EXIST;  // record already exists
  wrtRec(key, data, true);
  return dbResult;
}

/** save a record, overwrite if key exist else add a record **/
DB_RESULT simpleDB::writeRec(const char *key, const char *data) {
  if (openDB() != DB_OK) return dbResult;
  if (strlen(data) > dtW) {dbResult = MAX_CHARS; return dbResult;}  // data: too many characters
  if (findRec(key) > REC_EXIST) return dbResult;  // is there a find error?
  wrtRec(key, data, (dbResult ==  REC_EXIST ? false : true));
  return dbResult;
}

/** PRIVATE function: over write or add a record **/
DB_RESULT simpleDB::wrtRec(const char *key, const char *data, boolean eof) {
  dbResult = DB_OK;
  if (eof) if (!dbF.seek(dbF.size())) {dbResult = SEEK_ERR; return dbResult;}  // if record not exists then seek to EOF
  // format key and tombstone marker "key|0|" 
  char n[dbW + 1];
  sprintf(n, hlK, key);
  dbF.print(n);  // write key + tombstone to db
  // format data with trailing white spaces
  sprintf(n, hlR, data);
  dbF.print(n);    // write fields to db
  dbF.flush();     // save record physically to the SD
  if (eof) {dbR++; dbT++;}  // update record nr
  return dbResult;
}

/** read the record with the given "key", it is in record buffer dbB **/
DB_RESULT simpleDB::readRec(const char *key) {
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (findRec(key) != REC_EXIST) return dbResult;  // find position in file
  if (!dbF.seek(dbF.position() + nxW)) {dbResult = SEEK_ERR; return dbResult;}  // seek to first position of data (the '|' after tombstone marker)
  uint8_t len = dbF.readBytes(dbB, dtW);  //  get only the data, data in buffer dbB
  dbB[len] = '\0';
  if (len != dtW) dbResult = CORRUPT;
  return dbResult;
}

/** read a record on position "nr", it is in record buffer dbB, included tombstone records **/
DB_RESULT simpleDB::getRec(uint8_t nr) {  // nr: position of record
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (dbF.size() < dbW) {dbResult = DB_EMPTY; return dbResult;}  // db is empty
  if (!dbF.seek(nr * dbW)) {dbB[0] = '\0'; dbResult = SEEK_ERR; return dbResult;}
  uint8_t len = dbF.readBytes(dbB, dbW);  // record in buffer dbB
  dbB[len] = '\0';
  if (dbB[tsM] != '0') dbResult = TOMBSTONE; // if it is a tombstone
  if (len != dbW) dbResult = CORRUPT;
  return dbResult;
}

/** find record with the given "key", slow but simple and fast enough for a small db **/
DB_RESULT simpleDB::findRec(const char *key) {
  // seek to record, sequentieel search
  // false: not found, true: found
  if (strlen(key) != kyW) {dbResult = MAX_CHARS; return dbResult;}
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (!dbF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}  
  if (dbF.size() < dbW) {dbResult = DB_EMPTY; return dbResult;}  // db is empty
  uint16_t len;
  while (dbF.available()) {
   len = dbF.readBytes(nxB, nxW);
   nxB[len] = '\0';
   if (len != nxW) {dbResult = CORRUPT; break;}
   if (nxB[tsM] == '0') {  // not a tombstone?
     boolean f = true;
     for (uint8_t i = 0; i < kyW; i++) {
       if (nxB[i] != key[i]) {f = false; break;}
     }
     if (f) {
       if (!dbF.seek(dbF.position() - nxW, SeekSet)) break;  // seek to the start byte of the record
       dbResult = REC_EXIST;
       return dbResult;  // key is found
     }
   }
   if (!dbF.seek(dbF.position() + dtW)) break;  // skip data, set entree to next record
  }
  return NOT_FOUND;
}

/** tombstone a record, remove tombstones with comprDb **/
DB_RESULT simpleDB::delRec(const char *key) {
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (findRec(key) != REC_EXIST) return dbResult;  // find position in file
  if (!dbF.seek(dbF.position() + tsM)) {dbResult = SEEK_ERR; return dbResult;}
  dbF.print('1');
  dbF.flush();
  dbR--;  // update record nr
  return dbResult;
}

/** copy db, then remove tombstone records **/
DB_RESULT simpleDB::cmprDb() { 
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  dbResult = DB_OK;
  dbF.close();
  if (openDB() != DB_OK) return dbResult;
  if (SD.exists(cpN)) if (!SD.remove(cpN)) return REMOVE_ERR;  // delete file error
  cpF = SD.open(cpN, FILE_WRITE);
  if (!cpF) {dbResult = OPEN_ERR; return dbResult;}
  cpF = SD.open(cpN, FILE_WRITE);
  if (!cpF) {dbResult = OPEN_ERR; return dbResult;}
  // make a copy
  if (!dbF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}
  while (dbF.available()) cpF.print((char)dbF.read());
  cpF.flush();
  dbF.close();
  // delete db
  if (!SD.remove(dbN)) return REMOVE_ERR;  // delete file error

  // make a new db file
  if (openDB() != DB_OK) return dbResult;
  if (dbF.size() != 0) return EXIST_ERR;  // file not earased - is not possible?

  // copy the 'not as tombstone marked' records from copy to database
  uint16_t len;
  int16_t nr = 0;
  if (!cpF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}
  while (cpF.available()) {
   len = cpF.readBytes(dbB, dbW);
   dbB[len] = '\0';
   if (strlen(dbB) != len) {
     cpF.close();
     dbResult = CORRUPT;
     break;  // db corrupt
   }
   if (dbB[tsM] == '0') {  // if it is not a tombstone ...
     dbF.print(dbB);  // ... then copy it to the tmp file
     dbF.flush();
     nr++;
   }
  }
  cpF.close();  // do not remove, it is your backup
  dbR = nr;
  if (nr != dbR) return DIF_NR;  // difference in number of records
  return dbResult;
}
