#ifndef SIMPLEDB_SD_H
#define SIMPLEDB_SD_H

#include <Arduino.h>
#include <SPI.h>
#include <SD.h>

enum DB_RESULT {  // result code
 DB_OK       =  0,  // OK, no error
 DB_EMPTY    =  1,  // no record(s) in db
 NOT_FOUND   =  2,  // record not found
 REC_EXIST   =  3,  // record already exists
 /** ERRORs **/
 DIF_NR      =  4,  // difference in number of records (DB and backup DB)
 CORRUPT     =  5,  // db corrupt
 MAX_CHARS   =  6,  // key or data: too many characters
 MAX_REC     =  7,  // max records in db
 SEEK_ERR    =  8,  // file seek error  
 EXIST_ERR   =  9,  // no file
 BACKUP_ERR  = 10,  // backup problem
 FORMAT_ERR  = 11,  // format problem
 TOMBSTONE   = 12,  // deleted record
 REMOVE_ERR  = 13,  // delete file error
 RENAME_ERR  = 14,  // rename file error
 MOUNT_ERR   = 15,  // can not mount  
 OPEN_ERR    = 16,  // file open error  
 DB_ERR      = 17,  // db error
};

class simpleDB {
 private:
   DB_RESULT wrtRec(const char *key, const char *data, boolean eof);  // PRIVATE function: over write or add a record 

   uint8_t  kyW,        // FIXED key width
            tsM,        // pos. tombstone marker
            nxW,        // fixed index width (key width + tombstone marker "key|0|")
            dbW,        // record width, key|t|dddd|...|ddd (t=tombstone, ddd..=data, |=delimiter)
            ssP;        // Chip Select pi
   uint16_t mxR;        // max number of records
   char     dbN[32],    // database file  // short notation max 8.3 'naaaaame.ext'
            cpN[32],    // database back-up file
            hlK[7],     // strcpy(), holder key format string
            hlR[7],     // strcpy(), holder record format string
            *nxB;       // index read buffer
 public:
   uint8_t  dtW;        // max data width in a record, exclusive 'key' and 'tombstone' marker
   uint16_t dbR,        // record counter, exlusive tombstone records
            dbT;        // record counter, inclusive tombstone records
   char     *dbB;       // record read buffer

   DB_RESULT dbResult;  // DB_RESULT error code holder

   File dbF, cpF;  // DB and DB-backup 

   /** init an new or existing database **/
   DB_RESULT initDB(const char *nameDB, const char *backupDB, uint8_t ssPin, uint8_t keyLen, uint8_t recLen, uint16_t maxRec);  // SD
   
   /** simpleDB functions **/
   DB_RESULT nrRec();                                      // set counter dbR, the nummer of records, tombstone records are excluded
   DB_RESULT addRec(const char *key, const char *data);    // add a record with a UNIQUE key 
   DB_RESULT writeRec(const char *key, const char *data);  // save a record, overwrite if key exist else add a record 
   DB_RESULT readRec(const char *key);                     // read the record with the given "key", it is in record buffer dbB 
   DB_RESULT getRec(uint8_t nr);                           // read a record on position "nr", it is in record buffer dbB 
   DB_RESULT findRec(const char *key);                     // find record with the given "key", slow but simple and fast enough for a small db 
   DB_RESULT delRec(const char *key);                      // tombstone a record, remove tombstones with comprDb 
   DB_RESULT cmprDb();                                     // copy db, then remove tombstone records 
   DB_RESULT openDB();                                     // test if open, else open or create DB
   void      closeDB();                                    // close db
};

#endif
