#ifndef SIMPLE_DB_H
#define SIMPLE_DB_H

#include <simpleDB_SPIFFS.h>

/** simpleDB record format:
 *  "key.|T|field1.......|field2.......|field3.......|field4.......|field5.......|field6.......|"
 *  key = fixed length of 4 chars!!!
 *  | = fields delimiter
 *  T = tombstone flag. 0: record in use, 1: record forever unsusable. Delete?: use function "cmprDb()" (DB backup and compress)
 *  fields len = min. 1 to max. field<n>Len (standard 13)
 *  PLEASE NOTE: empty fields must contain a white space
 */

enum {
  mode              =  0,  // 0/1/2/3/4/5/6/7/8                0  // Man/Auto/Lux/Temp/Rise/Set/Sun/Night/Day
  pause             =  1,  // 0/1/2 no/pause/stop              0
  hh1               =  2,  // time 1 hh                        0
  mm1               =  3,  // time 1 mm                        0
  hh2               =  4,  // time 2 hh                        0
  mm2               =  5,  // time 2 mm                        0
  lux1              =  6,  // lux 1                          350
  lux2              =  7,  // lux 2                           60
  tmp1              =  8,  // temperature 1                   21.5
  tmp2              =  9,  // temperature 2                   21.0
  code              = 10,
  period            = 11,
  status            = 12,
  address           = 13,
  groupBit          = 14,
  unit              = 15,
  dimLevelPresent   = 16,
  dimLevel          = 17,  //                                 -1
  pin               = 18,
  repetitions       = 19,  //                                  2
  duration          = 20,  //                                500
  reversed          = 21,  //                                  0
  riseOffset        = 22,  // OFF minutes after sunRise       15
  setOffset         = 23,  // ON minutes before sunSet        15
  yoyo              = 24,  // anti-yo-yo time in minutes      10
  badLux            = 25,  // lux (< luxUp)                   40
  pin1              = 26,  // ON(/OFF) pin                     6   or ADS1115 module (0-15), or ...Steps arg1
  pin2              = 27,  // OFF(NULL) pin                    7   or ...Steps arg2
  counter1          = 28,  // mode 6: anti-yoyo minutes       -1  (free to use in other modes)
  counter2          = 29,  // mode 6: bad-weather minutes     -1  (free to use in other modes)
  counter3          = 30,  // mode 6: anti-yoyo minutes       -1  (free to use in other modes)
};  


class sHS_DB {
  private:
    /** define the user db record and fields parameters **/
    const char   nameDB[14] = "/node.db",   // if using SPIFFS: the name must start with a slash '/'
               backupDB[14] = "/node.cdb";  // backup to SD or SPIFFS
    
    const static uint8_t keyLen = 4,
                  fieldLen = 14,           // max. length value fields = 13 + 1
                  enumLen = counter3 + 1;  // eNum counter3 (records)
    const uint16_t recLen = 128,           // fixed stream length
                   maxRec = 32;            // max number of records
   
    /** parse a record to record[] **/
    void parseRec();
    
  public:  
    simpleDB db;         // db
    char record[enumLen][fieldLen],  // list with set/parsed record values
         recKey[keyLen + 1];         // to get a key from getRec(nr)
         
    /** open or create a DB **/
    void dbInit();
    /** get and parse a record to record[] **/
    void readRec(char * Key);
    /** get and parse a record to record[] **/
    void getRec(uint8_t nr);
    /** save a record from record[] **/
    void writeRec(char * Key);
    /** basic record **/
    void newRecord(); 

};
/** the list of simpleDB funtions
 *  =============================
    ** specific simpleDB funtions **
    uint8_t dtW;   // max data width in a record, exclusive 'key' and 'tombstone' marker
    uint16_t dbR;  // record counter, exlusive tombstone records
             dbT;  // record counter, inlusive tombstone records
    char *dbB;     // record read buffer

    ** init an new or existing database **
    DB_RESULT initDB(const char *nameDB, const char *backupDB, uint8_t keyLen, uint8_t recLen, uint16_t maxRec);
    ** simpleDB functions **
    DB_RESULT nrRec();                                      // set counter dbR, the nummer of records, tombstone records are excluded
    DB_RESULT addRec(const char *key, const char *data);    // add a record with a UNIQUE key 
    DB_RESULT writeRec(const char *key, const char *data);  // save a record, overwrite if key exist else add a record 
    DB_RESULT readRec(const char *key);                     // read the record with the given "key", it is in record buffer dbB 
    DB_RESULT getRec(uint8_t nr);                           // read a record on position "nr", it is in record buffer dbB 
    DB_RESULT findRec(const char *key);                     // find record with the given "key", slow but simple and fast enough for a small db 
    DB_RESULT delRec(const char *key);                      // tombstone a record, remove tombstones with comprDb 
    DB_RESULT cmprDb();                                     // copy db, then remove tombstone records 
    DB_RESULT openDB();                                     // test if open, else open or create DB
              closeDB();                                    // close db
    
    DB_RESULT dbResult;                                     // DB_RESULT error code holder
    
    enum DB_RESULT {  // result code
      DB_OK       =  0,  // OK, no error
      DB_EMPTY    =  1,  // no record(s) in db
      NOT_FOUND   =  2,  // record not found
      REC_EXIST   =  3,  // record already exists
      // ERRORs //
      DIF_NR      =  4,  // difference in number of records (DB and backup DB)
      CORRUPT     =  5,  // db corrupt
      MAX_CHARS   =  6,  // key or data: too many characters
      MAX_REC     =  7,  // max records in db
      SEEK_ERR    =  8,  // file seek error  
      EXIST_ERR   =  9,  // no file
      BACKUP_ERR  = 10,  // backup problem
      FORMAT_ERR  = 11,  // format problem
      TOMBSTONE   = 12,  // deleted record
      REMOVE_ERR  = 13,  // delete file error
      RENAME_ERR  = 14,  // rename file error
      MOUNT_ERR   = 15,  // can not mount  
      OPEN_ERR    = 16,  // file open error  
      DB_ERR      = 17,  // db error
};
**/  

#endif

