/***********************************************
 * DST/SUMMERTIME by Ewald Kendziorra sr, 2017 *
 ***********************************************/
// Code: from here and there ;=)

#include "simpleSummerTimeDST.h"

DST::DST(uint8_t _startMonth,     // 1 - 12 (jan - dec) 
         uint8_t _endMonth,       //   ...
         uint8_t _startMonthDay,  // 1 - 31, not used = 0 (if using ...WeekDay) 
         uint8_t _endMonthDay,    //   ...
         uint8_t _startWeekDay,   // 1 -  7 (Sun - Sat), not used = 0 (if using ...MonthDay)
         uint8_t _endWeekDay,     //   ...
         uint8_t _startHour,      // 0 - 23 (UTC)
         uint8_t _endHour,        //   ...
         uint8_t _startMinute,    // 0 - 59 (Lord Howe Island rules use a half-hour time zone ;=))
         uint8_t _endMinute,      //   ...
         int8_t  _startWeek,      // 1 -  4, last week of month: -1 (4 or 5)
         uint8_t _endWeek,        //   ... 
         uint16_t _dstOffset,     // dst/summertime offset in seconds
         boolean _hemisphere,     // Northern Hemisphere: true,  Southern Hemisphere: false
         double _timeZone) {      // UTC−12 to UTC+14 (Lord Howe Island rules use a half-hour time zone ;=))
  startMonth = _startMonth;
  endMonth =  _endMonth;
  startMonthDay = _startMonthDay;
  endMonthDay = _endMonthDay;
  startWeekDay = _startWeekDay;
  endWeekDay = _endWeekDay; 
  startHour = _startHour; 
  endHour = _endHour;
  startMinute = _startMinute;
  endMinute = _endMinute;
  startWeek = _startWeek;
  endWeek = _endWeek;
  dstOffset = _dstOffset;
  hemisphere = _hemisphere;
  timeZone = _timeZone;
}

void DST::calcDst() {
  calcDst(now());
}

void DST::calcDst(time_t t) {
  uint16_t y = year(t);
  /** fixed date? **/
  if (startMonthDay) {
    /** Month Day > 0: it's a fixed DST **/
    dstStart = calcTimeStamp(startHour, startMinute, 0, startMonthDay, startMonth, y);
    dstEnd = calcTimeStamp(endHour, endMinute, 0, endMonthDay, endMonth, y);
  }
  /** flex date DST **/
  /** last week 4/5 of the month? **/
  else {
    /** start **/
    if (startWeek == -1) {  // last week
      dstStart = calcTimeStamp(startHour, startMinute, 0, 1, startMonth + 1, y);  // get the first day of the next mondth
      dstStart -= oneDay;  // set last day of the month
      while (startWeekDay != weekday(dstStart)) dstStart -= oneDay;
    } 
    else {  // week 1-4
      dstStart = calcTimeStamp(startHour, startMinute, 0, 1, startMonth, y);  // get the first day of the next mondth
      dstStart += (startWeek - 1) * oneWeek;  // set the week of the month
      while (startWeekDay != weekday(dstStart)) dstStart += oneDay;
    }
    if (endWeek == -1) {
      /** end **/   
      dstEnd = calcTimeStamp(endHour, endMinute, 0, 1, endMonth + 1, y);  // get the first day of the next mondth
      dstEnd -= oneDay;  // set last day of the month
      while (endWeekDay != weekday(dstEnd)) dstEnd -= oneDay;
    }
    else {
      /** end **/
      dstEnd = calcTimeStamp(endHour, endMinute, 0, 1, endMonth, y);  // get the first day of the next mondth
      dstEnd += (endWeek - 1) * oneWeek;  // set the week of the month
      while (endWeekDay != weekday(dstEnd)) dstEnd += oneDay;
    }
  }
}

boolean DST::isItDst(time_t t, boolean calc) {
  if (calc) calcDst(t);
  /** location north or south of the equator? **/
  if (hemisphere) {
    if (t >= dstStart && t < dstEnd)       return 1;  // north
  }
  else if (!(t < dstStart && t >= dstEnd)) return 1;  // south
  return 0;
}

time_t DST::localTime(time_t t, boolean calc) {
  return t + (uint32_t)(timeZone * oneHour + (isItDst(t, calc) ? dstOffset : 0)); 
}

time_t DST::calcTimeStamp(uint8_t h, uint8_t m, uint8_t s, uint8_t D, uint8_t M, uint16_t Y) {  // get UNIX time_t Time Stamp from hh:mm:ss DD-MM-YY
  // UNIX timestamp = number of seconds since 1 Jan 00:00:00 1970
  if (Y < 100) Y += 2000;
  time_t t = 0;
  for (uint16_t i = 1970; i < Y; i++) t += yearDay(31, 12, i) * oneDay;
  return t + (yearDay(D, M, Y) - 1) * oneDay + (uint32_t)h * oneHour + (uint32_t)m * 60 + s;
}

uint16_t DST::yearDay(time_t t) {  // year days from timestamp
  return yearDay(day(t), month(t), year(t));
}

uint16_t DST::yearDay(uint8_t d, uint8_t m, uint16_t y) {  // year days from date numbers
  int monthdays[13] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
  return monthdays[m - 1] + (monthdays[m - 1] > 31 && leapYear(y) ? 1 : 0) + d;
}

boolean DST::leapYear(uint16_t y) {  // leap year?
  uint16_t Y = y < 100 ? y + 2000 : y;
  return ((Y % 4 == 0 && Y % 100 != 0) || Y % 400 == 0);
}

