/********************************************
 * Sunrise/set by Ewald Kendziorra sr, 2017 *
 ********************************************/

/** Source:
 *  Almanac for Computers, 1990
 *  published by Nautical Almanac Office
 *  United States Naval Observatory
 *  Washington, DC 20392
 *  Adapted by Ewald Kendziorra sr, www.kendziorra.nl
 **/

#include "simpleSunRiseSet.h"

sunRiseSet::sunRiseSet(uint16_t _dstOffset,  // dst/summertime offset in seconds
                       float _lat,           // The latitude is preceded by a minus sign ( – ) if it is south of the equator (a positive number implies north)
                       float _lon,           // the longitude is preceded by a minus sign if it is west of the prime meridian (a positive number implies east)
                       double _timeZone,     // UTC−12 to UTC+14 (Lord Howe Island rules use a half-hour time zone ;=))
                       float _zenith) {      //    offical      = 90.883 degrees (90° 50')
                                             //    civil        = 96 degrees
                                             //    nautical     = 102 degrees
                                             //    astronomical = 108 degrees
  dstOffset = _dstOffset;
  lat = _lat;
  lon =  _lon;
  timeZone =  _timeZone;
  zenith = _zenith;
}

/** today **/
void sunRiseSet::calcSun(boolean Dst) {
  calcSun(now(), Dst);
}

/** from date **/
void sunRiseSet::calcSun(time_t t, boolean Dst) {
  /** calc yearDay **/
  uint16_t N = floor(275 * month(t) / 9) - (floor((month(t) + 9) / 12) * (1 + floor((year(t) - 4 * floor(year(t) / 4) + 2) / 3)))  + day(t) - 30;

  /** get SunRise time **/
  float sunRiseDec = sunTime(N, true);
  sunRiseDec += timeZone + (Dst ? (dstOffset / 3600) : 0);  // to local time
  sunRiseHour = sunRiseDec;  // byte<-float ;=))
  sunRiseMinute = (sunRiseDec - sunRiseHour) * 60;

  /** get SunSet time **/
  float sunSetDec = sunTime(N, false);
  sunSetDec += timeZone + (Dst ? (dstOffset / 3600) : 0);  // to local time
  sunSetHour = sunSetDec;
  sunSetMinute = (sunSetDec - sunSetHour) * 60;
}

/** calc decimal time **/
float sunRiseSet::sunTime(uint16_t N, boolean rise) {
  float lngHour = lon / 15.0;
  float t =  N + ((rise ? 6.0 : 18.0) - lngHour) / 24.0;
  float M = (0.9856 * t) - 3.289;
  float L = M + (1.916 * sin(M * toRad)) + (0.020 * sin(2 * M * toRad)) + 282.634;
  L += (L < 0 ? 360 : L > 360 ? -360 : 0); 
  float RA = atan(0.91764 * tan(L * toRad)) * toDeg;
  RA += (RA < 0 ? 360 : RA > 360 ? -360 : 0);
  float Lquadrant = (floor(L / 90.0)) * 90.0;
  float RAquadrant = (floor(RA / 90.0)) * 90.0;
  RA = (RA + Lquadrant - RAquadrant) / 15.0;
  float sinDec = 0.39782 * sin(L * toRad);
  float cosDec = cos(asin(sinDec));
  float cosH = (cos(zenith * toRad) - (sinDec * sin(lat * toRad))) / (cosDec * cos(lat * toRad));
  float H = (rise ? (360.0 - acos(cosH) * toDeg) : (acos(cosH) * toDeg)) / 15; 
  float T = H + RA - (0.06571 * t) - 6.622;
  float UT = T - lngHour;
  UT +=  (UT < 0 ? 24 : UT > 24 ? -24 : 0);
  return UT;  // UTC decimal time 
}
